from datetime import datetime
from enum import Enum
from typing import List, Optional, Union

from odp.util.cheapdantic import BaseModel, Field


class PayloadType(str, Enum):
    DATASET = "dataset"
    DATA_COLLECTION = "data_collection"


class AutoCompletionResultType(str, Enum):
    DATASET = "dataset"
    DATA_COLLECTION = "data_collection"
    TAG = "tag"
    CONTRIBUTOR = "contributor"
    REGION = "region"
    LICENSE = "license"


class SearchScore(BaseModel):
    full_text_search_score: Optional[float] = None
    exact_match_score: Optional[float] = None
    trigram_match_score: Optional[float] = None
    levenshtein_match_score: Optional[float] = None
    total_score: Optional[float] = None


class DatasetSearchResultPayload(BaseModel):
    id: str
    title: str
    title_highlight: str
    description: str
    description_highlight: str
    tags: List[str] = Field(default_factory=list)
    tags_highlight: str


class CollectionSearchResultPayload(BaseModel):
    id: str
    title: str
    title_highlight: str
    description: str
    description_highlight: str


class ResultItem(BaseModel):
    id: str
    title: str
    description: str
    highlights: List[str] = Field(default_factory=list)
    match_type: str
    payload_type: PayloadType
    score: SearchScore
    payload: Union[DatasetSearchResultPayload, CollectionSearchResultPayload, dict]


class SearchResult(BaseModel):
    total: int
    top_score: Optional[SearchScore] = None
    match: bool
    results: List[ResultItem] = Field(default_factory=list)


# --- Search Request ---


class SearchFilter(BaseModel):
    tags: List[str] = Field(default_factory=list)
    types: List[PayloadType] = Field(default_factory=list)
    contributor_id: Optional[str] = None
    license_id: Optional[str] = None
    from_time: Optional[datetime] = None
    to_time: Optional[datetime] = None
    within: Optional[str] = None


class SearchRequest(BaseModel):
    query: str
    size: Optional[int] = None
    page: Optional[int] = None
    filter: Optional[SearchFilter]


class AutoCompletionScore(BaseModel):
    rank: float
    similarity: float
    edit_distance: int


class AutoCompletionResultItem(BaseModel):
    id: str
    type: AutoCompletionResultType
    title: str
    description: str
    score: AutoCompletionScore


class AutoCompletionResult(BaseModel):
    total: int
    top_score: Optional[AutoCompletionScore] = None
    results: List[AutoCompletionResultItem] = Field(default_factory=list)


class AutoCompletionGroupedResult(BaseModel):
    group: AutoCompletionResultType
    result: AutoCompletionResult


class AutoCompletionRequest(BaseModel):
    query: str
    size: Optional[int] = None



import io
import json
import logging
import typing
from typing import TYPE_CHECKING, Dict, Iterator, Optional, Union

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from odp.tabular_v2.util import Iter2Reader
from odp.util.util import size2human

if TYPE_CHECKING:
    from odp.tabular_v2.client.table import Table

read_retry_strategy = Retry(
    total=8,
    status_forcelist=[429, 502, 503, 504],
    allowed_methods=["HEAD", "GET", "OPTIONS", "POST"],
    backoff_factor=1,  # exponential sleep (1s, 2s, 4s…)
)

write_retry_strategy = Retry(
    total=8,
    status_forcelist=[429],
    allowed_methods=["HEAD", "GET", "OPTIONS", "POST"],
    backoff_factor=2,
)


class Client:
    def __init__(self, base_url: str):
        self._base_url = base_url

    class Response:
        # Abstraction for response object, shared between http client and test client
        def __init__(self, res: Union[requests.Response, Iterator[bytes], Dict, bytes]):
            if isinstance(res, requests.Response):
                if res.status_code == 204:
                    raise FileNotFoundError(res.text)
                if res.status_code == 400:
                    raise ValueError(res.text)
                res.raise_for_status()
            # logging.debug("response: %s", res)
            self.res = res

        def reader(self):
            if isinstance(self.res, bytes):
                return io.BytesIO(self.res)
            if isinstance(self.res, Iterator):
                return Iter2Reader(self.res)
            if isinstance(self.res, requests.Response):
                if self.res.status_code == 204:
                    raise FileNotFoundError(self.res.text)
                self.res.raise_for_status()
                if self.res.raw is None:
                    return io.BytesIO(self.res.content)
            if isinstance(self.res, dict):
                return io.BytesIO(bytes(str(self.res), "utf-8"))
            if isinstance(self.res, io.IOBase):
                return self.res
            if isinstance(self.res, requests.Response):
                return self.res.raw
            if isinstance(self.res, list):  # assuming list of bytes
                return Iter2Reader(iter(self.res))
            raise ValueError(f"unexpected type {type(self.res)}")

        def iter(self) -> Iterator[bytes]:
            if isinstance(self.res, bytes):
                return iter([self.res])
            if isinstance(self.res, Iterator):
                return self.res
            if isinstance(self.res, requests.Response):
                if self.res.status_code == 204:
                    raise FileNotFoundError(self.res.text)
                self.res.raise_for_status()
                if self.res.raw is None:
                    return iter([self.res.content])
                return self.res.raw
            if isinstance(self.res, dict):
                return iter([bytes(str(self.res), "utf-8")])
            raise ValueError(f"unexpected type {type(self.res)}")

        def all(self) -> bytes:
            if isinstance(self.res, bytes):
                return self.res
            if isinstance(self.res, Iterator):
                return b"".join(self.res)
            if isinstance(self.res, requests.Response):
                if self.res.status_code == 204:
                    raise FileNotFoundError(self.res.text)
                self.res.raise_for_status()
                if self.res.raw is None:
                    return self.res.content
                return self.res.raw.read()
            if isinstance(self.res, dict):
                return bytes(str(self.res), "utf-8")
            raise ValueError(f"unexpected type {type(self.res)}")

        def json(self) -> typing.Union[dict, list, None]:
            if isinstance(self.res, requests.Response):
                return self.res.json()
            if self.res is None:
                return None
            if isinstance(self.res, dict):
                return self.res
            if isinstance(self.res, list):
                return self.res
            if isinstance(self.res, bytes):
                return json.loads(self.res.decode("utf-8"))
            if isinstance(self.res, requests.Response):
                if self.res.status_code == 204:
                    raise FileNotFoundError(self.res.text)
                self.res.raise_for_status()
                if self.res.raw is None:
                    return None
                return json.loads(self.res.raw.read().decode("utf-8"))
            raise ValueError(f"unexpected type {type(self.res)}")

    def _request(
        self,
        path: str,
        data: Union[Dict, bytes, Iterator[bytes], io.IOBase, None] = None,
        params: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        retry: bool = True,  # if it is safe to retry the request on failure
    ) -> Response:
        if retry:
            retry_strategy = read_retry_strategy  # read operations are safe to retry in many cases
        else:
            retry_strategy = write_retry_strategy  # more restrictive

        with requests.Session() as s:
            adapter = HTTPAdapter(pool_connections=4, pool_maxsize=4, max_retries=retry_strategy)
            s.mount("http://", adapter)
            s.mount("https://", adapter)
            if isinstance(data, dict):
                logging.debug("sending %s %s %s", path, params, data)
                res = s.post(self._base_url + path, headers=headers, params=params, json=data, stream=True)
            elif isinstance(data, bytes):
                logging.debug("sending %s %s (%s)", path, params, size2human(len(data)) if data else "-")
                res = s.post(self._base_url + path, headers=headers, params=params, data=data, stream=True)
            elif isinstance(data, io.IOBase):
                logging.debug("sending file-like %s %s", path, params)
                res = s.post(self._base_url + path, headers=headers, params=params, data=data, stream=True)
            elif isinstance(data, Iterator):
                logging.debug("sending %s %s iterator...", path, params)
                res = s.post(self._base_url + path, headers=headers, params=params, data=data, stream=True)
            elif data is None:
                logging.debug("sending %s %s no data", path, params)
                res = s.post(self._base_url + path, headers=headers, params=params, stream=True)
            else:
                raise ValueError(f"unexpected type {type(data)}")
        clen = res.headers.get("Content-Length", "")
        if not clen:
            clen = res.headers.get("Transfer-Encoding", "")
        logging.debug("got response %s (%s) for %s", res.status_code, clen, path)
        return self.Response(res)

    def table(self, table_id: str) -> "Table":
        """
        create a table handler for the given table_id
        @param table_id:
        @return:
        """
        from odp.tabular_v2.client import Table

        return Table(self, table_id)



import urllib.parse
from io import BytesIO
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import requests

from ..util.cheapdantic import BaseModel
from .dto.file_dto import FileMetadataDto
from .exc import OdpFileAlreadyExistsError, OdpFileNotFoundError, OdpValidationError
from .http_client import OdpHttpClient
from odp.dto import DatasetDto


class OdpRawStorageClient(BaseModel):
    http_client: OdpHttpClient
    raw_storage_endpoint: str = "/data"

    @property
    def raw_storage_url(self) -> str:
        """The URL of the raw storage endpoint, including the base URL.

        Returns:
            The raw storage URL
        """
        return f"{self.http_client.base_url}{self.raw_storage_endpoint}"

    def _construct_url(self, resource_dto: DatasetDto, endpoint: str = "") -> str:
        if resource_dto.metadata.uuid:
            return f"{self.raw_storage_url}/{resource_dto.metadata.uuid}{endpoint}"
        else:
            return f"{self.raw_storage_url}/catalog.hubocean.io/dataset/{resource_dto.metadata.name}{endpoint}"

    @staticmethod
    def _encode_filename(filename: str) -> str:
        filename = filename.lstrip("/")
        return urllib.parse.quote(filename, safe="")

    def get_file_metadata(self, resource_dto: DatasetDto, file_metadata_dto: FileMetadataDto) -> FileMetadataDto:
        """Get file metadata by reference.

        Args:
            resource_dto: Dataset manifest
            file_metadata_dto: File metadata to retrieve

        Returns:
            The metadata of the file corresponding to the reference
                requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

        Raises:
            OdpFileNotFoundError: If the file does not exist
        """
        url = self._construct_url(resource_dto, endpoint=f"/{self._encode_filename(file_metadata_dto.name)}/metadata")

        response = self.http_client.get(url)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpFileNotFoundError(f"File not found: {file_metadata_dto.name}") from e
            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

        return FileMetadataDto(**response.json())

    def list(
        self, resource_dto: DatasetDto, metadata_filter: Optional[Dict[str, Any]] = None
    ) -> Iterable[FileMetadataDto]:
        """List all files in a dataset.

        Args:
            resource_dto: Dataset manifest
            metadata_filter: List filter

        Returns:
            List of files in the dataset
        """

        cursor_next = None
        while True:
            page, cursor_next = self.list_paginated(resource_dto, metadata_filter=metadata_filter, cursor=cursor_next)
            yield from page
            if not cursor_next:
                break

    def list_paginated(
        self,
        resource_dto: DatasetDto,
        metadata_filter: Optional[Dict[str, Any]] = None,
        cursor: Optional[str] = None,
        page_size: int = 1000,
        limit: int = 0,
    ) -> Tuple[List[FileMetadataDto], str]:
        """List page

        Args:
            resource_dto: Dataset manifest
            metadata_filter: List filter
            cursor: Optional cursor for pagination
            limit: Optional #This argument will be deprecated, you should use page_size instead.
                This is used for backward compatibility. limit will be ignored.
            page_size: Optional limit for each page

        Returns:
            Page of return values
        """

        url = self._construct_url(resource_dto, endpoint="/list")
        params = {}

        if cursor:
            params["page"] = cursor
        if page_size:
            params["page_size"] = page_size
        if limit:
            from warnings import warn

            warn(
                "limit argument will be deprecated, you should use page_size instead", DeprecationWarning, stacklevel=2
            )

        response = self.http_client.post(url, params=params, content=metadata_filter)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise OdpValidationError("API argument error") from e
            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

        content = response.json()
        return [FileMetadataDto(**item) for item in content["results"]], content.get("next")

    def upload_file(
        self,
        resource_dto: DatasetDto,
        file_metadata_dto: FileMetadataDto,
        contents: Union[bytes, BytesIO],
        overwrite: bool = False,
    ) -> FileMetadataDto:
        """Upload data to a file.

        Args:
            resource_dto: Dataset manifest
            file_metadata_dto: File metadata
            contents: File contents
            overwrite: Overwrite file if it exists

        Returns:
            The metadata of the uploaded file
        """
        url = self._construct_url(resource_dto, endpoint=f"/{self._encode_filename(file_metadata_dto.name)}")

        if isinstance(contents, bytes):
            contents = BytesIO(contents)

        # Ensure contents is in the correct format (bytes)
        if isinstance(contents, BytesIO):
            contents = contents.read()

        headers = {"Content-Type": "application/octet-stream"}

        response = self.http_client.patch(url, params={"overwrite": overwrite}, headers=headers, content=contents)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpFileNotFoundError(f"File not found: {file_metadata_dto.name}") from e
            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

        return self.get_file_metadata(resource_dto, file_metadata_dto)

    def create_file(
        self,
        resource_dto: DatasetDto,
        file_metadata_dto: FileMetadataDto,
        contents: Union[bytes, BytesIO, None] = None,
    ) -> FileMetadataDto:
        """Create a new file.

        Args:
            resource_dto: Dataset manifest
            file_metadata_dto: File metadata
            contents: File contents

        Returns:
            The metadata of the newly created file
        """

        url = self._construct_url(resource_dto)
        headers = {"Content-Type": "application/json"}
        response = self.http_client.post(
            url,
            headers=headers,
            content=file_metadata_dto.json(exclude_unset=True),
        )

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise OdpValidationError("API argument error") from e
            elif response.status_code == 409:
                raise OdpFileAlreadyExistsError(f"File already exists: {file_metadata_dto.name}")
            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

        file_meta = FileMetadataDto(**response.json())

        if contents:
            return self.upload_file(resource_dto, file_meta, contents)

        return self.get_file_metadata(resource_dto, file_meta)

    def download_file(
        self,
        resource_dto: DatasetDto,
        file_metadata_dto: FileMetadataDto,
        save_path: Optional[str] = None,
    ):
        """Download a file.

        Args:
            resource_dto: Dataset manifest
            file_metadata_dto: File metadata of file
            save_path: File path to save the downloaded file to
        """
        url = self._construct_url(resource_dto, endpoint=f"/{self._encode_filename(file_metadata_dto.name)}")

        response = self.http_client.get(url)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpFileNotFoundError(f"File not found: {file_metadata_dto.name}") from e
            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

        if save_path:
            with open(save_path, "wb") as file:
                file.write(response.content)
        else:
            return response.content

    def delete_file(self, resource_dto: DatasetDto, file_metadata_dto: FileMetadataDto):
        """Delete a file. Raises exception if any issues.

        Args:
            resource_dto: Dataset manifest
            file_metadata_dto: File metadata og file to delete.

        Returns:
            `True` if the file was deleted, `False` otherwise
        """
        url = self._construct_url(resource_dto, endpoint=f"/{self._encode_filename(file_metadata_dto.name)}")

        response = self.http_client.delete(url)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpFileNotFoundError(f"File not found: {file_metadata_dto.name}") from e

            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")



"""This module contains the set of ODP SDK exceptions."""


class OdpError(Exception):
    """Base class for exceptions in this module."""


class OdpAuthError(OdpError):
    """Exception raised for authentication errors."""


class OdpUnauthorizedError(OdpError):
    """Exception raised for unauthorized requests."""


class OdpForbiddenError(OdpError):
    """Exception raised for forbidden requests."""


class OdpTokenValidationError(OdpError):
    """Exception raised for invalid tokens."""


class OdpResourceNotFoundError(OdpError):
    """Exception raised when a resource is not found."""


class OdpResourceExistsError(OdpError):
    """Exception raised when a resource already exists."""


class OdpValidationError(OdpError):
    """Exception raised when a resource is not found."""


class OdpFileNotFoundError(OdpError):
    """Exception raised when a file is not found."""


class OdpFileAlreadyExistsError(OdpError):
    """File already exists"""


class OpenTableStageInvalidAction(OdpError):
    """Exception when table is getting deleted and it has active sessions."""



from typing import IO, Any, Callable, Dict, List, Optional, Protocol, Type, Union

JsonType = Union[None, int, str, bool, List["JsonType"], Dict[str, "JsonType"]]


class JsonParser(Protocol):
    """JSON serialization/deserialization interface"""

    @staticmethod
    def load(
        fp,
        decoded_type: Type,
        cast_decimal: bool = True,
        cls: Optional[Type] = None,
        parse_float: Optional[Callable[[str], float]] = None,
        parse_int: Optional[Callable[[str], int]] = None,
        parse_constant: Optional[Callable[[str], JsonType]] = None,
        **kwargs
    ) -> JsonType:
        ...

    @staticmethod
    def loads(
        s: str,
        cast_decimal: bool = True,
        cls: Optional[Type] = None,
        parse_float: Optional[Callable[[str], float]] = None,
        parse_int: Optional[Callable[[str], int]] = None,
        parse_constant: Optional[Callable[[str], JsonType]] = None,
        **kwargs
    ) -> JsonType:
        ...

    @staticmethod
    def dump(
        obj: Any,
        fp: IO,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: Optional[Type] = None,
        indent=None,
        separators=None,
        default=None,
        sort_keys=False,
        **kwargs
    ):
        ...

    @staticmethod
    def dumps(
        obj: Any,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: Optional[Type] = None,
        indent: Optional[int] = None,
        separators: Optional[str] = None,
        default: Optional[Callable[[str], str]] = None,
        sort_keys: bool = False,
        **kwargs
    ) -> str:
        ...



from datetime import datetime, timedelta
from typing import List, Literal, Optional, Union
from uuid import UUID, uuid4

from odp.util.cheapdantic import BaseModel


class TablePartitioningSpec(BaseModel):
    columns: List[str]
    transformer_name: str
    args: Optional[List[Union[int, float, str]]] = None

    def serialize(self) -> bytes:
        return self.json().encode("utf-8")


class TableStage(BaseModel):
    stage_id: UUID
    status: Literal["active", "commit", "commit-failed", "delete"]
    created_time: datetime
    expiry_time: datetime
    updated_time: Optional[datetime] = None

    error: Optional[str] = None
    error_info: Optional[dict] = None

    def serialize(self) -> bytes:
        return self.json().encode("utf-8")

    @classmethod
    def generate(cls, expiry_time: timedelta) -> "TableStage":
        now = datetime.now()

        return cls(stage_id=uuid4(), status="active", created_time=now, expiry_time=now + expiry_time)

    def dict(self, **kwargs) -> "DictStrAny":  # noqa: F821
        exclude_unset = kwargs.pop("exclude_unset", True)
        return super().dict(exclude_unset=exclude_unset, **kwargs)



from datetime import datetime
from typing import Optional
from uuid import UUID

from odp.util.cheapdantic import BaseModel


class ResourceStatus(BaseModel):
    """Resource status model"""

    num_updates: int = 0
    """Number of time the manifest has been updated"""

    created_time: datetime
    """Created timestamp"""

    created_by: UUID
    """UUID of user that created the resource"""

    updated_time: datetime
    """Last updated timestamp"""

    updated_by: UUID
    """UUID of user that updated the resource"""

    deleted_time: Optional[datetime] = None
    """Deleted timestamp - used for soft-delete"""

    deleted_by: Optional[UUID] = None
    """UUID of user that deleted the resource"""

    @classmethod
    def create(cls, created_by: UUID) -> "ResourceStatus":
        t = datetime.now()
        return ResourceStatus(created_time=t, created_by=created_by, updated_time=t, updated_by=created_by)

    class Config:
        extra = "allow"



import re
from math import log2

IEC_UNITS = ["KiB", "MiB", "GiB", "TiB"]


def size2human(size: int) -> str:
    if size == 0:
        return "0B"
    p = int(log2(size) // 10.0)

    if p < 1:
        return f"{size}B"
    if p > len(IEC_UNITS):
        p = len(IEC_UNITS)
    converted_size = size / 1024**p
    return f"{converted_size:.1f}{IEC_UNITS[p - 1]}"


_VALID = re.compile(r"^[a-zA-Z0-9_.-]+$")


def safe_id(s: str, empty: bool = False, max_len: int = 100) -> str:
    """
    check if the id is safe, raise an exception if not
    """
    if s is None:
        s = ""
    s = str(s)
    if s == "" and not empty:
        raise ValueError("id is empty")
    if not s:
        return ""
    if len(s) > max_len:
        raise ValueError(f"id is too long: {len(s)} > {max_len}")
    if not _VALID.match(s):
        raise ValueError(f"id contains invalid characters: {repr(s)}")
    return s



from typing import Dict, Optional, Union

from odp.client.auth import TokenProvider
from odp.tabular_v2.client import Client


class ClientAuthorization(Client):
    def __init__(self, base_url, token_provider: TokenProvider):
        super().__init__(base_url)
        self.token_provider = token_provider

    def _request(
        self,
        path: str,
        data: Union[Dict, bytes, None] = None,
        params: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        retry: bool = True,
    ) -> Client.Response:
        headers = headers or {}
        headers["Authorization"] = self.token_provider.get_token()
        headers["user-agent"] = self.token_provider.user_agent
        return super()._request(path, data, params, headers, retry)



from .data_collection import DataCollectionDto, DataCollectionSpec
from .dataset import Attribute, Citation, DatasetDto, DatasetSpec
from .distribution import Distribution
from .observable import ObservableDto, ObservableSpec
from odp.dto.resource import ResourceDto

del ResourceDto
